# mill.py, Markdown interface for llama.cpp
# Copyright (C) 2024 unworriedsafari <unworriedsafari@tilde.club>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>

"""
# README

`mill.py v1.1.1`

Markdown interface for [llama.cpp](//github.com/ggerganov/llama.cpp).


## Requirements

1. [Python 3.x](//python.org) (tested on `3.11`)
2. [llama.cpp](//github.com/ggerganov/llama.cpp) (tested on `b1860`)

Developed and tested on Linux. I believe it should also work on Windows or Mac.


## Features

1. Lets you interact with `llama.cpp` using Markdown
2. Enables you to use almost every `llama.cpp` option
3. Makes no assumptions about what model you want to use
4. Lets you change any option at any point in the document
5. Caches prompts automatically
6. Streams output
7. Runs in a CLI environment as well as a CGI environment
8. Reads input document from `stdin`, writes output document to `stdout`
9. Lets you add support for any other language or LLM through Python modules


## Example

Contents of `hello.md`:

    ## Variables

    ```mill-llm
    --model
    mixtral-8x7b-instruct-v0.1.Q5_0.gguf
    ```
    
    ```mill-llm
    --ctx-size
    0
    ```
    
    ```mill-llm
    --keep
    -1
    ```
    
    ```mill
    message template
    

    Me:
    
    > [INST]  [/INST]
    
    Bot:
    
    >
    ```
    
    ```mill
    prompt indent
    >
    ```
    
    ## Chat
    
    ```mill
    prompt start
    ```
    
    Me:
    
    > [INST] Hello, how are you? [/INST]
    
    Bot:
    
    >

Command:

```bash
export MILL_LLAMACPP_MAIN=path/to/llama.cpp/main /path/to/mill_cli.py <hello.md
```

Output:

    ## Variables

    ```mill-llm
    --model
    mixtral-8x7b-instruct-v0.1.Q5_0.gguf
    ```
    
    ```mill-llm
    --ctx-size
    0
    ```
    
    ```mill-llm
    --keep
    -1
    ```
    
    ```mill
    message template


    Me:

    > [INST]  [/INST]

    Bot:
    
    >
    ```
    
    ```mill
    prompt indent
    >
    ```
    
    ## Chat
    
    ```mill
    prompt start
    ```
    
    Me:
    
    > [INST] Hello, how are you? [/INST]
    
    Bot:
    
    > Hello! I'm just a computer program, so I don't have feelings, but I'm here to help you with any questions you have to the best of my ability. Is there something specific you would like to know or talk about?</s>
    
    Me:
    
    > [INST]  [/INST]
    
    Bot:
    
    >


## Adding support for other languages

Markdown support is included. To add another language:

1. Create a new Python module named `mill_lang_language_id` where all
   non-alphanumeric characters of `language_id` are replaced by underscores.
2. Implement a `parse` function similar to the one in `mill_lang_markdown.py`.
3. Put your module anywhere on the Python path of `mill.py`.
4. When using the CLI interface, pass the `-l language_id` argument.
5. When using the CGI interface, pass the `language=language_id` query-string
   parameter.


## Adding support for other LLMs

`llama.cpp` support is included. Adding support for LLMs is similar to adding
support for other languages:

1. Create a new Python module named `mill_llm_llm_id` where all
   non-alphanumeric characters of the `llm_id` part are replaced by
   underscores.
2. Implement a `generate` function similar to the one in
   `mill_llm_llama_cpp.py`.
3. Put your module anywhere on the Python path of `mill.py`.
4. When using the CLI interface, pass the `-e llm_id` argument.
5. When using the CGI interface, pass the `llm_engine=llm_id` query-string
   parameter.
"""

import importlib, re


def load_module(name):
    module = 'mill_' + re.sub(r'\W', '_', name).lower()
    return importlib.import_module(module)
    

def main(language, llm_engine, input_lines):
    with language.parse(input_lines) as p:
        if p.returncode != 0:
            return p.returncode

        with llm_engine.generate(p.llm_vars, p.prompt) as g:
            word = ''
            while True:
                char = g.read(1)
                word += char

                if not char:
                    if word:
                        p.print_generated_text(word)
                        word = ''
                    break

                if re.match(r'\W', char):
                    p.print_generated_text(word)
                    word = ''

            if g.returncode == 0:
                p.print_message_template()

            return g.returncode
