#! /usr/bin/env python
#
# mill.py, Markdown interface for llama.cpp
# Copyright (C) 2024 unworriedsafari <unworriedsafari@tilde.club>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>

r"""
## CLI install + usage

1. Clone the Git repo or download these files:
    1. `mill_cli.py`
    2. `mill.py`
    3. `mill_lang_markdown.py`
    4. `mill_llm_llama_cpp.py`
    5. `mill_example_markdown_llama_cpp.py`
2. Put files 2-5 on the Python path of `mill_cli.py`. Easy solution: put all
   files in the same folder.
3. Set the environment variable `MILL_LLAMACPP_MAIN` to the path of
   `llama.cpp/main` or your wrapper around it.
4. Pipe your Markdown document to `mill_cli.py`.

```bash
export MILL_LLAMACPP_MAIN=/path/to/llama.cpp/main
python /path/to/mill_cli.py <document.md
```

The result printed on stdout is the original document with the generated text
from the LLM added.

Like any CLI tool, you can also use it through SSH:

```bash
cat document.md | ssh <host> \
    "MILL_LLAMACPP_MAIN=/path/to/llama.cpp/main python /path/to/mill_cli.py" \
    2>/dev/null
```

Use the command-line arguments to select a different language or LLM. You can
use `-h` for a usage description.
"""

import argparse, mill, sys


if __name__ == '__main__':
    argparser = argparse.ArgumentParser(
        description='Markdown interface for llama.cpp')

    argparser.add_argument('-r', '--readme', action='store_true',
                           help='display the README')
    argparser.add_argument('-l', '--language', default='markdown',
                           help='language')
    argparser.add_argument('-e', '--llm-engine', default='llama.cpp',
                           help='LLM engine')

    args = argparser.parse_args()
    
    language = mill.load_module(f'lang_{args.language}')
    llm_engine = mill.load_module(f'llm_{args.llm_engine}')
    input_lines = sys.stdin.readlines()

    if args.readme or not ''.join([line.strip() for line in input_lines]):
        print(mill.__doc__.strip())
        print()
        print()
        print(__doc__.strip())
        print()
        print()
        print(language.__doc__.strip())
        print()
        print()
        print(llm_engine.__doc__.strip())
        try:
            example = \
                mill.load_module(f'example_{args.language}_{args.llm_engine}')
            print()
            print()
            print(example.__doc__.strip())
        except ModuleNotFoundError:
            pass
        exit(0)
    
    exit_code = mill.main(language, llm_engine, input_lines)

    exit(exit_code)
